import os
import warnings
from functools import reduce
from typing import List, Union, Optional, Dict

from . import markdown
from .page import Page, RootPage
from .pagesfile import PagesFile
from .options import Options


class PageNotFoundError(Exception):
    def __init__(self, page: str, context: str = None):
        message = 'Page "{page}" not found.'
        if context:
            message += ' [{context}]'

        super().__init__(message.format(page=page, context=context))


class TitleInRootPagesFileWarning(Warning):
    pass


class Factory:

    INDEX_PAGE_NAME = 'index'

    def __init__(self, options: Options):
        self.options = options

    def create(self, config: List) -> RootPage:
        """ Creates a root page containing the whole tree of pages from the mkdocs config """
        children = self._create_pages(config)

        pages_file = self._load_pages_file('')
        if pages_file.title is not None:
            warnings.warn(
                'Using the "title" attribute in the {filename} file of the doc root has no effect'
                    .format(filename=self.options.filename),
                TitleInRootPagesFileWarning)

        children = self.arrange_pages(children, pages_file)

        return RootPage(children)

    def create_page(self, config_entry: Union[Dict, str]) -> Page:
        """ Creates a page from an entry in the mkdocs config """
        if isinstance(config_entry, str):
            return self._create_leaf_page(None, config_entry)

        # The config dictionary always contains one entry, retrieve it
        [title, value] = next(iter(config_entry.items()))

        if isinstance(value, str):
            return self._create_leaf_page(title, value)
        else:
            return self._create_branch_page(title, value)

    def _create_pages(self, config: List) -> List[Page]:
        """ Calls create_page for every entry in the config list """
        return [self.create_page(c) for c in config]

    def _create_leaf_page(self, title: Optional[str], path: str) -> Page:
        """ Creates a leaf page, a page pointing to an actual markdown file """
        try:
            with open(path) as file:
                title = markdown.extract_h1(file)
        except FileNotFoundError:
            pass

        return Page(title, path)

    def _create_branch_page(self, title: str, config_children: list) -> Page:
        """ Creates a branch page, a directory page with child pages """
        children = self._create_pages(config_children)
        path = self.common_dirname(children)

        pages_file = self._load_pages_file(path)
        title = pages_file.title or title

        children = self.arrange_pages(children, pages_file)

        return Page(title, path, children)

    def _load_pages_file(self, path: str):
        """ Loads and parses the pages file for a given path """
        try:
            if path is None:
                raise FileNotFoundError
            return PagesFile.load_from(os.path.join(path, self.options.filename))
        except FileNotFoundError:
            # Default to empty pages file
            return PagesFile()

    def common_dirname(self, children: List[Page]) -> Optional[str]:
        """ Determines the common dirname of all given pages """
        if children:
            return reduce(self._common_dirname, children, children[0].dirname)

    def _common_dirname(self, path: Optional[str], page: Page) -> Optional[str]:
        """ Reduces pages to a common dirname """
        if page and path is not None and page.dirname == path:
            return path

    def arrange_pages(self, pages: List[Page], pages_file: PagesFile) -> List[Page]:
        """ Sorts the given pages based on the arrange configuration from the pages file """
        pages_by_basename = self._group_pages_by_basename(pages)
        arranged_pages = set()
        rest_index = None
        result = []

        # Add pages from arrange configuration
        for index, path in enumerate(pages_file.arrange):
            if path == PagesFile.ARRANGE_REST_TOKEN:
                rest_index = index
            elif path in pages_by_basename:
                matching_pages = pages_by_basename[path]
                arranged_pages.update(matching_pages)
                result.extend(matching_pages)
            else:
                raise PageNotFoundError(path, pages_file.path)

        if rest_index is None:
            # If no rest token is used in the arrange configuration, add remaining pages at the end
            rest_index = len(result)

        # Add remaining pages
        for page in pages:
            if page in arranged_pages:
                # Skip pages that have already been added
                continue

            if (
                not pages_file.arrange
                and not self.options.disable_auto_arrange_index
                and page.basename is not None
                and os.path.splitext(page.basename)[0] == self.INDEX_PAGE_NAME
            ):
                # Automatically position index file at the beginning
                result.insert(0, page)
            else:
                result.insert(rest_index, page)

            rest_index += 1

        return result

    def _group_pages_by_basename(self, pages: List[Page]) -> Dict[str, List[Page]]:
        """ Creates a dictionary with mapping basenames to a list of matching pages """
        result = {}
        for page in pages:
            if page.basename in result:
                result[page.basename].append(page)
            else:
                result[page.basename] = [page]
        return result



import os
from typing import List, Optional, Union, Dict


class Page:

    def __init__(self, title: Optional[str], path: Optional[str], children: Optional[List['Page']] = None):
        self.title = title
        self.path = path
        self.dirname = os.path.dirname(path) if path else None
        self.basename = os.path.basename(path) if path else None
        self.children = children or []

    def _children_to_mkdocs(self, collapse_single_children: bool = False) -> List:
        return list(map(lambda child: child.to_mkdocs(collapse_single_children), self.children))

    def to_mkdocs(self, collapse_single_pages: bool = False) -> Union[List, Dict, str]:
        if self.children:
            children = self._children_to_mkdocs(collapse_single_pages)
            if collapse_single_pages and len(children) == 1:
                return children[0]
            else:
                return {
                    self.title: children
                }
        elif self.title:
            return {
                self.title: self.path
            }
        else:
            return self.path


class RootPage(Page):

    def __init__(self, children: List[Page]):
        super().__init__(None, '', children)

    def to_mkdocs(self, collapse_single_pages: bool = False) -> List:
        return self._children_to_mkdocs(collapse_single_pages)



import yaml
from typing import Optional, List


class DuplicateRestTokenError(Exception):
    def __init__(self, context: str = None):
        message = 'Rest token "..." is only allowed once'
        if context:
            message += ' [{context}]'

        super().__init__(message.format(context=context))


class PagesFile:

    TITLE_ATTRIBUTE = 'title'
    ARRANGE_ATTRIBUTE = 'arrange'
    ARRANGE_REST_TOKEN = '...'

    def __init__(self, *, title: Optional[str] = None, arrange: Optional[List[str]] = None, path: Optional[str] = None):
        self.title = title
        self.arrange = arrange or []
        self.path = path

    @staticmethod
    def load_from(path: str) -> 'PagesFile':
        with open(path) as file:
            contents = yaml.load(file) or {}
            title = contents.get(PagesFile.TITLE_ATTRIBUTE)
            arrange = contents.get(PagesFile.ARRANGE_ATTRIBUTE)

            if title is not None:
                if not isinstance(title, str):
                    raise TypeError(
                        'Expected "{attribute}" attribute to be a string - got {type} [{context}]'
                        .format(attribute=PagesFile.TITLE_ATTRIBUTE, type=type(title), context=path)
                    )
            if arrange is not None:
                if not isinstance(arrange, list) or not all(isinstance(s, str) for s in arrange):
                    raise TypeError(
                        'Expected "arrange" attribute to be a list of strings - got {type} [{context}]'
                        .format(attribute=PagesFile.ARRANGE_ATTRIBUTE, type=type(arrange), context=path)
                    )
                if arrange.count(PagesFile.ARRANGE_REST_TOKEN) > 1:
                    raise DuplicateRestTokenError(path)

            return PagesFile(title=title, arrange=arrange, path=path)



class Options:
    def __init__(self, *, filename: str, disable_auto_arrange_index: bool, collapse_single_pages: bool):
        self.filename = filename
        self.disable_auto_arrange_index = disable_auto_arrange_index
        self.collapse_single_pages = collapse_single_pages



from mkdocs import utils as mkdocs_utils
from mkdocs.config import config_options
from mkdocs.plugins import BasePlugin

from . import utils
from .factory import Factory
from .options import Options


class AwesomePagesPlugin(BasePlugin):

    DEFAULT_PAGES_FILENAME = '.pages'

    config_scheme = (
        ('filename', config_options.Type(mkdocs_utils.string_types, default=DEFAULT_PAGES_FILENAME)),
        ('disable_auto_arrange_index', config_options.Type(bool, default=False)),
        ('collapse_single_pages', config_options.Type(bool, default=False))
    )

    def on_config(self, config):

        with utils.cd(config['docs_dir']):
            options = Options(**self.config)
            pages = Factory(options).create(config['pages'])
            config['pages'] = pages.to_mkdocs(options.collapse_single_pages)

        return config



from setuptools import setup, find_packages


setup(
    name='mkdocs-awesome-pages-plugin',
    version='1.1.1',
    description='An MkDocs plugin that simplifies configuring page titles and their order',
    long_description='The awesome-pages plugin allows you to customize how your pages show up the navigation of your '
                     'MkDocs without having to configure the full structure in your ``mkdocs.yml``. It extracts the '
                     'title from your markdown files and gives you detailed control using a small configuration file '
                     'directly placed in the relevant directory of your documentation. See `Github '
                     '<https://github.com/lukasgeiter/mkdocs-awesome-pages-plugin>`_ or the README.md for more '
                     'details.',
    keywords='mkdocs python markdown wiki',
    url='https://github.com/lukasgeiter/mkdocs-awesome-pages-plugin/',
    author='Lukas Geiter',
    author_email='info@lukasgeiter.com',
    license='MIT',
    python_requires='>=3.5',
    install_requires=[
        'mkdocs>=0.17'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    packages=find_packages(exclude=['*.tests']),
    entry_points={
        'mkdocs.plugins': [
            'awesome-pages = mkdocs_awesome_pages_plugin.plugin:AwesomePagesPlugin'
        ]
    }
)



import re
from typing import TextIO, Optional


def extract_h1(file: TextIO) -> Optional[str]:
    for line in file:
        match = re.match('#([^#].*)', line)
        if match:
            return match.group(1).strip()



import os


class cd:
    """ Context manager for changing the current working directory """

    def __init__(self, new_path):
        self.new_path = os.path.expanduser(new_path)

    def __enter__(self):
        self.saved_path = os.getcwd()
        os.chdir(self.new_path)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.saved_path)
